/**
 * An entity is the basic unit of shared state.
 */
export interface Entity<T = any> {
    /** Unique identifier for the entity */
    name: string;
    /**
     * Sets the entity to its initial value.
     *
     * This is not normally invoked directly by the app, but can be
     * useful for unit testing.
     */
    init(): void;
    /**
     * Gets the current value of the entity.
     *
     * @returns the entity value
     */
    get(): T;
    /**
     * Updates the value of the entity.
     *
     * @param valueOrUpdaterFn - new value or an updater function
     * @param alias - optional name for the update
     */
    set(valueOrUpdaterFn: T | ((value: T) => T), alias?: string): void;
    /**
     * Subscribes to entity updates, and returns an unsubscribe function.
     *
     * This is not normally invoked directly by a React component (use the
     * hook instead), but can be useful inside other functions.
     */
    subscribe(subscriberFn: (newValue: T) => any): () => void;
    /**
     * Binds the entity value to component state. This is shorthand for the
     * `useEntity(thisEntity)` hook.
     *
     * @param transform - optional data transformation function
     * @param equality - optional custom equality function
     * @returns the current entity value
     */
    use<C = T>(transform?: (value: T) => C, equality?: (a: C, b: C) => boolean): C;
}
/**
 * A plug-in extends the behavior of an entity by composing on top of its
 * original methods. Plug-ins can be stacked on top of another.
 */
export type Plugin = {
    [K in keyof Entity as Exclude<K, 'name'>]?: (origMethod: Entity[K], entity: Entity) => Entity[K];
};
/**
 * Creates and returns a new entity.
 *
 * Initial value can be _any_ type, including primitives, but cannot be
 * explicitly `undefined` (prefer `null` for this).
 *
 * It can also be a `Promise` (unresolved, i.e. do not `await`), in which
 * case its value is implicitly `undefined` until the Promise resolves.
 *
 * @param initialValue - required default value
 * @param pluginsOrAlias - optional array list of plug-ins (or alias shorthand)
 * @returns the entity object
 */
export declare function entity<T = any>(initialValue: Promise<T>, pluginsOrAlias?: Plugin[] | string): Entity<T | undefined>;
export declare function entity<T = any>(initialValue: T, pluginsOrAlias?: Plugin[] | string): Entity<T>;
